// >>>>> base.cc

#ifndef _BASE_CC_
#define _BASE_CC_

/* Common structures and I/O functions 
 */
  
#include "cfg.h"
#include <os.cc>
#include <time.h>
#include <nothreads.h>
#include "unicode.cc"
#include "tab.cc"

void bailout() {exit(1);}


#ifdef VERBOSE
#define VERB(x) if(stdmsg) { x ; fflush(stdmsg); }
#else
#define VERB(x)
#endif

#ifdef STATISTICS
#define STAT(x) x
#else
#define STAT(x)
#endif

static inline F R(F x) { return x*ulog2(x+F_tiny(F(1))); }
static inline F Q(F a, F b) { return R(a+b)-R(a)-R(b); }

typedef double D;  //don't change this since it is used with printf()

struct tree{
  union{int l; int minlevel;};  /* left and right indexes of the classes (or words) contained in current class */
  union{int r; int maxlevel;};  /* or if it is a word then draw_tree() uses it for level information */
  int level; /* level is used for progress logging in mmi_clustering_x.cc */
};           /* it is a level of a node above the leafs */

struct mmi_result{
  int A;           /* A=c.A, all words have indexes  1:(A-1) */
                   /*   CAUTION: load_dictionary() returns A-1 */
  int B;           /* B number of unique bigrams in the input text */
  int C;           /* C=c.C,  C < c.A-1, number of words selected for classification */
  tree *forest;    /* [2*A-1] indexes 1:(A-1) represent words, while c.A:(c.A+c.C-2) represent classes */
  F I;             /* original mutual information */
  FF *loss;        /* [c.C-1]  0:(c.C-2) contains information loss in k-th step. */
};

FILE* stdmsg=0;    /* standard message output (usually stderr) */


union lebe_cast{
  u32 x;
  u8 y[4];
};
static inline int put8(FILE* f, u8 byte) { return putc(byte, f); }
static inline int put32(FILE* f, u32 x)
{
  lebe_cast l; l.x=x; int ec=0;
  for(int k=0; k<4; k++) ec|=put8(f,l.y[k]);
  return ec;
}

static inline int same_file(char* a, char* b)
{
  struct stat sb0, sb1;
  if (lstat(a,&sb0)>=0 && lstat(b,&sb1)>=0) 
    return (sb0.st_dev==sb1.st_dev && sb0.st_ino==sb1.st_ino);
  else return 0;
}

static FILE* mkdir_fopen(char* name, char* flags, int mode) UNUSED;
static FILE* mkdir_fopen(char* name, char* flags, int mode)
{
  FILE* fd=fopen(name, flags);
  if(fd) return fd;
  char* ptr=strrchr(name, '/');
  if(!ptr) return 0;
  *ptr=0;
  if(rec_mkdir(name, mode|((mode&0222)>>1)|0700)<0) return 0;
  *ptr='/';
  return fopen(name, flags);
} 

static int mkdir_open(char* name, int flags, int mode) UNUSED;
static int mkdir_open(char* name, int flags, int mode)
{
  int fd=open(name, flags, mode);
  if(fd>=0) return fd;
  char* ptr=strrchr(name, '/');
  if(!ptr) return -1;
  *ptr=0;
  if(rec_mkdir(name, mode|((mode&0222)>>1)|0700)<0) return -1;
  *ptr='/';
  return open(name, flags, mode);
}


/* Save machine representation of the tree
 */
void save_tree(char* fname, char* dirname, mmi_result& r)
{
  FILE* f=mkdir_fopen(fname, "w", 0644);
  if(!f) error("save_tree(): Cant open output file %s (%m)", fname); 
  char* ap; 
  abs_path(dirname, ap);
  fprintf(f,"%s\n", ap); 
  free(ap);
  fprintf(f,"%d %d\n", r.A-1, r.C-1);
  for(int k=r.A; k<r.A+r.C-1; k++)
    fprintf(f,"%d %d\n", r.forest[k].l, r.forest[k].r);
  fclose(f);
}


/* In all following blocks-related functions A is the number of words in the
 * dictionary (the number returned by load_dictionary()) and abc is
 * dictionary itself having dummy word at [0] and actual words at [1] to [A]
 *
 * Format of eqmap is an array [0:A] using indexes [1:A] to map each word
 * from the dictionary to its class number (class numbers start from 1 and
 * occupy a contiguous interval). Format of blocks is an array of length less
 * than 2*A. It contains blocks of positive numbers separated by zeroes all
 * of this being ended by -1. Each block represents class of words. In
 * normalized form (obtained by normalize_blocks()), sigleton classes are
 * removed to save space (and processing time in some circumstances).
 *
 * eqmap2blocks() always produces normalized blocks therefore it may happen
 * that blocks2eqmap(eqmap2blocks(x))!=x, since some blocks may become
 * renumbered. The following equalities are true, however.
 * blocks2eqmap(eqmap2blocks(blocks2eqmap(y)))=blocks2eqmap(y)
 * eqmap2blocks(blocks2eqmap(eqmap2blocks(x)))=eqmap2blocks(x)
 */
void save_blocks(char* fname, char** abc, int* blocks, FF* affinity)
{
  FILE *out=mkdir_fopen(fname, "w", 0644);
  if(!out) error("save_blocks(): cannot open %s (%m)", fname);
  for(;*blocks>=0; blocks++) 
    if(*blocks==0) fprintf(out, "\n"); 
    else{
      int l=fprintf(out,"%s", abc[*blocks]);
      if(l<0) error("save_blocks(): cannot write (%m)");
      if(affinity) fprintf(out,"%*s %10.6f\n", max(0,40-l), "", double(affinity[*blocks]));
      if(fprintf(out,"\n")<0) error("save_blocks(): cannot write (%m)");
    }
  if(fclose(out)<0) error("save_blocks(): cannot write (%m)");
}

/* Note that this can only be called when the hash table htab is not in use!
 * Since the only other place in mmi_clustering that does use htab is
 * text2bin.cc it is OK, at least for now. If affinity is !=0 then it will
 * be filled (must be already allocated to array [0:A] according to the
 * second column of the file (if not present than all will be zero).
 * Singleton words have affinity:=0.
 */
int* load_blocks(char* fname, char** abc, FF* affinity, int A)
{
  if(affinity) fill(affinity, FF(0), A+1);
  init_htab();
  for(int i=1; i<=A; i++){
    hash_ins(abc[i], hash_chain)->ct=i;
  }
  int* blocks; alloc(blocks,2*A);  
  FILE* in=fopen(fname,"r");
  if(!in) error("load_blocks(): cannot open %s for reading (%m)", fname);
  char symbol[2*PATH_MAX];
  int pos=0, la=getc(in);
  while(la>0){
    while(la==' ' || la=='\t' || la=='\r') la=getc(in);
    if(la!='\n'){
      int pp=0;
      while(!(la==' ' || la=='\t' || la=='\r' || la=='\n' || la<0)) 
        {if(pp<998) {symbol[pp++]=la;} la=getc(in);}
      symbol[pp]=0;
      while(la==' ' || la=='\t' || la=='\r') la=getc(in);
      double af=0;
      if(la>='0' && la<='9' || la=='+' || la=='-'){
        ungetc(la,in);
        if(fscanf(in,"%lg", &af)!=1) error("load_blocks(): number expected after symbol %s", symbol);
        la=getc(in);
      }
      while(la>0 && la!='\n') la=getc(in);
      la=getc(in);  //eat newline
      hnod* h=hash_qry(symbol);
//fprintf(stderr,"sym(%d)=%s\n", strlen(symbol), symbol);
      if(!h){
        fprintf(stderr,"WARNING: word `%s' from %s not in the dictionary\n", symbol, fname);
        continue;
      }
      if(h->ct<0) error("duplicate occurence of word `%s' found in %s", symbol, fname);
      blocks[pos++]=h->ct;
      if(affinity) affinity[h->ct]=af;
      h->ct=-1;
//printf("%s\n", symbol);
    }else{
//printf("-----------------\n");
      la=getc(in);
      if(pos>=2 && blocks[pos-2]==0) pos--;
      else blocks[pos++]=0;
    }
  }
  blocks[pos]=-1;
  fclose(in);
  done_htab();
//save_blocks("/dev/stdout", abc, blocks, 0);
  return blocks;
}

DECLARE_hsort(int*, (*high-*low))
FF* nhsort_affinity; int* nhsort_base;
DECLARE_nhsort(affinity_sort, int*, ({int t=(*high-*low); t?t:({int h=high-nhsort_base, l=low-nhsort_base; isgn(nhsort_affinity[l]-nhsort_affinity[h]);}); }))
int* eqmap2blocks(int* eqmap, FF* affinity, int A)
{
  int* ptr[A+1]; for(int i=0; i<=A; i++) ptr[i]=eqmap+i;
  if(!affinity)
    hsort(ptr+1, A);   //could be done even faster by radix sort :-)
  else{
    nhsort_base=eqmap;        //put words with higher affinity to class at the head of the block
    nhsort_affinity=affinity;
    affinity_sort(ptr+1, A);
  }
  int* blocks; alloc(blocks, 2*A);
  int ct=0;
  int last_class=*(ptr[1]);
  for(int i=1; i<=A; i++){
    int wclass=*(ptr[i]);
    int word=ptr[i]-eqmap;
    if(wclass!=last_class){
      if(ct>=2 && blocks[ct-2]==0) ct--;  //skip singletons
      else blocks[ct++]=0;
    }
    blocks[ct++]=word;
    last_class=wclass;
  }
  if(ct>=2 && blocks[ct-2]==0) ct--;
  blocks[ct]=-1;
  return blocks;
}

int* blocks2eqmap(int* blocks, int A)
{
  int* eqmap;
  alloc(eqmap, A+1); fill(eqmap,-1,A+1); 
  int ct=1;
  for(;*blocks>=0; blocks++) if(*blocks==0) ct++; else eqmap[*blocks]=ct;
  for(int i=1; i<=A; i++) if(eqmap[i]<0) eqmap[i]=ct++;
  return eqmap;
}

void save_eqmap(char* fname, char** abc, int* eqmap, FF* affinity, int A)
{
  int* blocks=eqmap2blocks(eqmap, affinity, A);
  save_blocks(fname, abc, blocks, affinity);
  free(blocks);
}

int* load_eqmap(char* fname, char** abc, FF* affinity, int A)
{
  int* blocks=load_blocks(fname, abc, affinity, A);
  int* em=blocks2eqmap(blocks, A);
  free(blocks);
  return em;
}

void normalize_blocks(int* blocks, FF* affinity)
{
  int A=0,k;
  for(k=0; blocks[k]>0; k++) A=max(blocks[k],A);
  int* eqmap=blocks2eqmap(blocks,A);
  int* newblocks=eqmap2blocks(eqmap,affinity,A);
  for(k=0; newblocks[k]>0; k++) blocks[k]=newblocks[k];
  free(eqmap); free(newblocks);
}

static int load_dictionary(char** &dictionary, char* dir);

/* Load mmi_result and its corresponding dictionary. dirname has to be
 * allocated before the call to 2048 characters (and will contian
 * bin-directory name upon return). Returns index of the root class.
 */
int load_tree(char *fname, mmi_result& r, char* dirname, char**&abc)
{
  int x, y, xx, yy;
  FILE* f=fopen(fname,"r");
  if(!f) error("load_tree(): Cant open file %s (%m)", fname);
  int line=1;
  if(fscanf(f," %2047s%*[^\n]\n", dirname)!=1)
    error("load_tree(): Cant scanf() the first line of %s (feof=%d)", fname, feof(f));
  line++;
  int A=load_dictionary(abc, dirname);
  if(fscanf(f," %d %d \n", &x,&y)!=2)
    error("load_tree(): %s:2: two numbers separated by a space expected (feof=%d)", fname, feof(f));  
  line++;
  if(A!=x) error("load_tree(): Incompatibile dictionary: A=%d but %s claims %d", A, fname, x);
  r.A=A+1;
  r.C=y+1;
  tree t000; t000.l=0; t000.r=0; t000.level=0;
  alloc(r.forest, 2*r.A-1); fill(r.forest, t000, 2*r.A-1);
  alloc(r.loss, r.C-1); fill(r.loss, FF(0), r.C-1);
    
  int M=r.A+r.C-2+1;
  for(int k=0; k<y; k++, line++){
    if(fscanf(f," %d %d \n", &xx,&yy)!=2)
      error("load_tree(): %s:%d: two numbers separated by a space expected (feof=%d)", fname, line, feof(f));
    if( !( 0<xx && xx<M && 0<yy && yy<M) ) error("load_tree(): %s:%d: value out of range [1,%d], xx=%d, yy=%d", M-1, xx,yy);
    r.forest[k+A+1].l=xx;
    r.forest[k+A+1].r=yy;
  }
  
  if(!feof(f)) error("load_tree(): %s:%d: EOF expected", fname, line);
  fclose(f);
  return M-1;
}

/* Print human readable representation of the tree, including losses
 */
void print_tree(FILE*f, int machine, mmi_result &res, char** abc)
{
  int A=res.A, C=res.C;  
  double ls=0;
  for(int k=0; k<C-1; k++) ls+=res.loss[k];
  if(!machine){
    fprintf(f,"total MI =             %19.16f\n", D(res.I));
    fprintf(f,"loss sum =             %19.16f\n", D(ls));
  }
  for(int k=A; k<A+C-1; k++){
    fprintf(f,"%5d    ", A+C-1-k);
    if(!machine) fprintf(f," %19.16f ", D(res.loss[k-A]));
    int w=res.forest[k].l;
    if(w<A) fprintf(f,".%-14s ", abc[w]);
    else fprintf(f,"^%-14d ", A+C-1-w);
    w=res.forest[k].r;
    if(w<A) fprintf(f,".%-14s ", abc[w]);
    else fprintf(f,"^%-14d ", A+C-1-w);
    fprintf(f,"\n");
  }
}

/* Print human readable representation of the tree, including losses
 */
void print_tree(FILE*f, mmi_result &res, char** abc){print_tree(f, 0, res, abc);}
void print_portable_tree(FILE* f, mmi_result &res, char** abc){print_tree(f, 1, res, abc);}

#if 0
/* Pozor ... tohle je tricky, sortovat se muze az teprv co jsem oznackoval
   levels, jinak by to psalo nesmyly. Proto se to zatim nepouziva
*/
char* sort_tree(mmi_result &res, char** abc, int node)
{
  if(node>=res.A) {
    char* left =sort_tree(res, abc, res.forest[node].l);
    char* right=sort_tree(res, abc, res.forest[node].r);  
    if(strcmp(left,right)>0){
      swap(res.forest[node].l, res.forest[node].r);
      swap(left, right);
    }
    return left;
  }else
    return abc[node];
}

#endif

#ifdef UTF8_GRAPH
  #define T_HORIZ    "━"
  #define T_TEE      "┳"
  #define T_VERT     "┃"
  #define T_CORNER   "\x9b"
  #define T_UP       "⬆"
  #define T_DOWN     "⬇"
  #define ustrlen(a) utf8_strlen(a)
#else
  #define ustrlen(a) strlen(a)
  #ifndef POOR_GRAPH
    #define T_HORIZ    ""
    #define T_TEE      ""
    #define T_VERT     ""
    #define T_CORNER   ""
    #define T_UP       ""
    #define T_DOWN     ""
  #else
    #define T_HORIZ    "-"
    #define T_TEE      "+"
    #define T_VERT     "|"
    #define T_CORNER   "'"
    #define T_UP       "^"
    #define T_DOWN     ","
  #endif
#endif

#define PRINT_STREE    1    //draw_tree options
#define PRINT_ILOSS    2
#define PRINT_LEVELS   4
#define PRINT_FREQS    8
#define PRINT_QS      16

void do_draw_tree(FILE* f, int options, int **eqmaps, FF** affnts, int blcknr, mmi_result &res, char** abc, i64* histogram, int* len, int* back, int depth, int p)
{
  back[depth]=p;
  if(p<res.A){
    int merge_time=-1; FF Iloss=0.0;
    char Iloss_quirk1=' ', Iloss_quirk2=' ';    
    for(int i=depth; i>0; i--){
      int q=back[i], q1=back[i-1];
      if(res.forest[q1].l!=q){      //merge_time (and levels) written along a given word
        merge_time=q1-res.A;        //refer to the internal node directly accesible
        Iloss=res.loss[merge_time]; //from that word's line thru T_CORNER (---')
        if(merge_time>0 && Iloss < res.loss[merge_time-1]) Iloss_quirk1='*';
        if(merge_time<res.C-2 && Iloss > res.loss[merge_time+1]) Iloss_quirk2='+';
        break;
      }
    }
    if(options & PRINT_ILOSS) fprintf(f,"%12.8f ",D(Iloss));
    if(options & PRINT_QS)    fprintf(f,"%c%c ", Iloss_quirk1, Iloss_quirk2);
    if(options & PRINT_LEVELS)fprintf(f,"%2d %2d %2d %2d ", depth, res.forest[p].level, res.forest[p].minlevel,res.forest[p].maxlevel); 
    if(options & PRINT_FREQS) fprintf(f,"%6d %10lld ", merge_time, histogram[p]);
    for(int k=0; k<blcknr; k++){
      fprintf(f,"%5d ", eqmaps[k][p]);
      if(affnts[k]) fprintf(f,"%+8.5f ", D(affnts[k][p]));
    }
    fprintf(f,"%s ", abc[p]);
    int i;
    for(i=depth; i>0 ; i--){
      int q=back[i], q1=back[i-1];
      for(int k=len[q1]-len[q]-1; k>0; k--) fprintf(f,T_HORIZ);
      if(res.forest[q1].l!=q) {fprintf(f,T_CORNER); break;}
      fprintf(f,T_TEE);
    }
    
    for(i--; i>0; i--){
      int q=back[i], q1=back[i-1];
      for(int k=len[q1]-len[q]-1; k>0; k--) fprintf(f," ");
      if (res.forest[q1].l==q) fprintf(f,T_VERT);
      else fprintf(f," ");
    }
    
    if(fprintf(f,"\n")<0) error("do_draw_tree(): cannot write (%m)");
  }
  else{
    do_draw_tree(f,options,eqmaps,affnts,blcknr,res,abc,histogram,len,back,depth+1, res.forest[p].l);
    do_draw_tree(f,options,eqmaps,affnts,blcknr,res,abc,histogram,len,back,depth+1, res.forest[p].r);
  }
}

int reach_word(mmi_result &r, int k)
{
  if(k<r.A) return k;
  k=r.forest[k].r;
  while(k>=r.A) k=r.forest[k].l;
  return k;
}

void determine_levels(mmi_result &r)
{
  for(int k=1; k<r.A; k++) r.forest[k].level=0;  
  int maxlevel=0;
  for(int k=r.A; k<r.A+r.C-1; k++){
    int a=r.forest[k].l, b=r.forest[k].r;
    int l=1+max(r.forest[a].level, r.forest[b].level);
    r.forest[k].level=l;
    if(maxlevel<l) maxlevel=l;
    int w=reach_word(r,k);
    r.forest[w].level=l; r.forest[w].maxlevel=maxlevel;
  }
  int minlevel=maxlevel;
  for(int k=r.A+r.C-2; k>=r.A; k--){
    int l=r.forest[k].level;
    if(minlevel>l) minlevel=l;
    int w=reach_word(r,k);
    r.forest[w].minlevel=minlevel;
  }
  int root=r.A+r.C-2;
  while(root>=r.A) root=r.forest[root].l;
  r.forest[root].maxlevel=maxlevel+1;
  r.forest[root].minlevel=maxlevel+1;
  r.forest[root].level=maxlevel+1;
}

void draw_tree_core(FILE*f, int options, mmi_result &res, char** abc, i64* histogram, int **eqmaps, FF** affnts, int blcknr)
{
  int len[res.A+res.C-1], back[res.C-1];
  for(int k=1; k<res.A; k++) len[k]=ustrlen(abc[k])+1;
  for(int k=res.A; k<res.A+res.C-1; k++) len[k]=1+max(len[res.forest[k].l], len[res.forest[k].r]);
  if(options & PRINT_ILOSS) fprintf(f,"%12s ", "MI_LOSS");
  if(options & PRINT_QS) fprintf(f,"Qq ");
  if(options & PRINT_LEVELS){
    fprintf(f,".D .L "T_DOWN"L "T_UP"L ");
    determine_levels(res);
  }
  if(options & PRINT_FREQS) fprintf(f,"%6s %10s ","ADDED", "FREQUENCY");
  FF* aff[blcknr];
  for(int p=0; p<blcknr; p++){
    int flg=0;
    if(affnts!=0) for(int k=1; k<res.A; k++) flg|=(affnts[p][k]!=0);
    aff[p] = flg ? affnts[p]:0;
    fprintf(f,"BLCK%d ", p);
    if(flg) fprintf(f,"AFFINITY%d ", p);
  }
  fprintf(f,"CLASSIFICATION TREE\n");
  do_draw_tree(f, options, eqmaps, aff, blcknr, res, abc, histogram, 
               len, back, 0, res.A+res.C-2); 
  if(fflush(f)<0) error("draw_tree(): cannot write (%m)");
}

void draw_tree(FILE*f, int options, mmi_result &res, char** abc, i64* histogram, int **eqmaps, FF** affnts, int blcknr)
{
  if(options & PRINT_STREE){
    int flags[res.A]; fill(flags,0,res.A);
    for(int k=res.A; k<res.A+res.C-1; k++){
      int idx=res.forest[k].l;
      if(idx<res.A) flags[idx]=1;
      idx=res.forest[k].r;
      if(idx<res.A) flags[idx]=1;
    }
    int L=1;
    for(int k=1; k<res.A; k++) if(flags[k]) L=max(L,int(strlen(abc[k]))+1);
    char* abc2[res.A];
    char base[L*res.A]; fill(base, ' ', L*res.A);
    fill(abc2, (char*)"!!-INTERNAL-ERROR-IN-draw_separable_tree()-!!", res.A);
    
    for(int k=1; k<res.A; k++) if(flags[k]) {
      abc2[k]=base+L*k;
      strcpy(abc2[k]+L-1-strlen(abc[k]), abc[k]);
    }
    draw_tree_core(f,options,res,abc2,histogram,eqmaps,affnts,blcknr);
  }else 
    draw_tree_core(f,options,res,abc,histogram,eqmaps,affnts,blcknr);
}


static int bin_load_file(u8*& bfr, int &bfr_len, char* directory, char* filename) UNUSED;
static int bin_load_file(u8*& bfr, int &bfr_len, char* directory, char* filename)
{
  int plen=strlen(directory);
  char path[plen+strlen(filename)+2];
  strcpy(path, directory); path[plen]='/'; strcpy(path+plen+1, filename);

  int fd=open(path, O_RDONLY);
  if(fd<0) error("Can't open file %s (%s)", path, strerror(errno));
  struct stat st;
  int r=lstat(path, &st); 
  if( r<0 || !S_ISREG(st.st_mode)) error("file %s is not regular", path); 
  
  int L=st.st_size;  
  if (L>bfr_len){
    if(bfr) free(bfr);
    alloc(bfr, L);
    bfr_len=L;
  }
  
  L=xread(fd,(char*)bfr,L);
    if(L<0) error("read error %s on file %s", strerror(errno), path);
  close(fd);
  return L;
}

template <class T> static inline int tbin_load_file(T*& bfr, int &bfr_len, char* directory, char* filename){
  u8* bfr0=(u8*)bfr;
  int ret=bin_load_file(bfr0,bfr_len,directory,filename)/sizeof(T);
  bfr=(T*)bfr0;
  return ret;
}

static inline int count_lines(u8* bfr, int L)
{
  int s=0;
  for(int k=0; k<L; k++) if(bfr[k]=='\n') s++;
  return s;
} 

/* Returns same L as load_dictionary(_,dir) would return
 */
int dictionary_size(char* dir)
{
  char f[strlen(dir)+100]; sprintf(f, "%s/histogram", dir);
  if(!file_exists(f)) error("file %s does not exist", f);
  return file_size(f)/sizeof(i64) - 1;
}

/*
Returns L, the number of entries in the dictionary where
dictionary[]={"TOTAL", "l_1",...,"l_L"}. Note that the zeroth ("TOTAL")
entry is not counted (histogram has total sum at its place therefore the
name). Function assumes newline after each entry in dictionary file
INCLUDING the LAST ONE.
*/
static int load_dictionary(char** &dictionary, char* dir)
{
  u8* bfr=0; int L=0, i,j; 
  int alph_bf_len=bin_load_file(bfr, L, dir, "dictionary");
  alloc(dictionary,2+count_lines(bfr,alph_bf_len));
  dictionary[0]="TOTAL";
  dictionary[1]=(char*)bfr;
  for(i=0,j=2; i<alph_bf_len; i++){
    if(bfr[i]=='\n'){
      bfr[i]=0; 
      dictionary[j++]=(char*)bfr+i+1;
    }
  }  
  return j-2;
}

/*
Loads histogram of occurences of the particular word in the corpus, where
hist[0]=#(total number of words). Returns number of words (same number as
load_dictionary())
*/
static int load_histogram(i64* &hist, char* dir) UNUSED;
static int load_histogram(i64* &hist, char* dir)
{
  u8* bfr=0; int L=0; 
  int r=bin_load_file(bfr, L, dir, "histogram");
  hist=(i64*)bfr;
  return (r/sizeof(i64))-1;
}

int pipe_pid=0;
int open_infile(char* pp, char* path)
{
  pipe_pid=0; //value signalling that we are not using pipe (used by bury_zombie())
  if(!file_exists(path)) error("file %s does not exist", path);
  int in;
  if(!pp){
    in=open(path, O_RDONLY);
    if(in<0) error("can't open file %s (%m)", path);
  }else{
    int pd[2];
    if(pipe(pd)<0) error("cannot create pipe (%m)");
    int redir[]={0, pd[1],2,-1};
    if((pipe_pid=run(".", 0, redir, pp,path,path,path,path,path,path,path,path,path,path,path,path,path,path,path,path,path,path,path,path))<0)
      error("cannot spawn %s (%m)", pp);
    in=pd[0];
    close(pd[1]);
  }
  return in;
}

FILE* fopen_infile(char* pp, char* path)
{ 
  int inf=open_infile(pp,path);
  FILE* in=fdopen(inf,"r");
  if(!in) error("cannot fdopen() pipe descriptor %d for reading (%m)", inf);
  return in;
}

void bury_zombie()
{
  if(pipe_pid){
    int s;
    waitpid(pipe_pid, &s, 0); 
    if(!WIFEXITED(s) || WEXITSTATUS(s)!=0)
      error("preprocessor exited abnormally, returning exit code %d", WEXITSTATUS(s));
  }
}

int count_ps(char* pp)
{
  int ct=0;
  if(pp) for(;*pp;pp++) if(*pp=='%') ct++;
  return ct;
}

void check_ps(char* pp)
{
  if(count_ps(pp)>20) error("too many occurences of `%%' in the preprocessor call");
}

#endif 
